/* 
 *    linux/drivers/video/fbcon-jpegdec.c - a tiny jpeg decoder.
 *      
 *      (w) August 2001 by Michael Schroeder, <mls@suse.de>
 *                  
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "jpegdecoder_nolibjpeg.h"
#include "jpeg_header.h"
#include "jconfig.h"

#define ISHIFT 11

#define IFIX(a) ((long)((a) * (1 << ISHIFT) + .5))
#define IMULT(a, b) (((a) * (b)) >> ISHIFT)
#define ITOINT(a) ((a) >> ISHIFT)

/* special markers */
#define M_BADHUFF	-1

struct in {
	unsigned char *p;
	unsigned int bits;
	int left;
	int marker;
};

/*********************************/
struct dec_hufftbl;
struct enc_hufftbl;

union hufftblp {
	struct dec_hufftbl *dhuff;
	struct enc_hufftbl *ehuff;
};

struct scan {
	int dc;			/* old dc value */

	union hufftblp hudc;
	union hufftblp huac;
	int next;		/* when to switch to next scan */

	int cid;		/* component id */
	int hv;			/* horiz/vert, copied from comp */
	int tq;			/* quant tbl, copied from comp */
};

/*********************************/

#define DECBITS 10		/* seems to be the optimum */

struct dec_hufftbl {
	int maxcode[17];
	int valptr[16];
	unsigned char vals[256];
	unsigned int llvals[1 << DECBITS];
};

inline static void decode_mcus (struct in *, int *, int, struct scan *, int *);
static void dec_makehuff (struct dec_hufftbl *, int *, unsigned char *);

/*********************************/

static void idctqtab (unsigned char *, int *);
static void idct (int *, int *, int *, long, int);
static void scaleidctqtab (int *, int);

/*********************************/

static void initcol (int[][64]);

static void col221111 (int *, unsigned char *, int);

/*********************************/
struct comp {
	int cid;
	int hv;
	int tq;
};

#define MAXCOMP 4
static struct comp comps[MAXCOMP] =
{
    { 0x01, 0x22, 0x00 },
    { 0x02, 0x11, 0x01 },
    { 0x03, 0x11, 0x01 },
    { 0x00, 0x00, 0x00 }
};

static struct scan dscans[MAXCOMP];

static unsigned char quant[4][64];

static struct dec_hufftbl dhuff[4];

#define dec_huffdc (dhuff + 0)
#define dec_huffac (dhuff + 2)

static struct in in;
	
int dquant[3][64];

void nolibjpeg_decode(unsigned char *pic, unsigned char * buf, int width, int height)
{
	int mcusx, mcusy, mx, my;
	int dcts[6 * 64 + 16];
	int out[64 * 6];
	int max[6];
	int i;

	if ((height & 15) || (width & 15))
		return;

	mcusx = width >> 4;
	mcusy = height >> 4;

	/* set input context */
	in.p = buf; 
	in.left = 0;
	in.bits = 0;
	in.marker = 0;

	/* for each component. Reset dc values. */
	for (i = 0; i < 3; i++)
		dscans[i].dc = 0;
	
	for (my = 0; my < mcusy; my++) {
		for (mx = 0; mx < mcusx; mx++) {
			decode_mcus(&in, dcts, 6, dscans, max);
			idct(dcts, out, dquant[0], IFIX(128.5), max[0]);
			idct(dcts + 64, out + 64, dquant[0], IFIX(128.5), max[1]);
			idct(dcts + 128, out + 128, dquant[0], IFIX(128.5), max[2]);
			idct(dcts + 192, out + 192, dquant[0], IFIX(128.5), max[3]);
			idct(dcts + 256, out + 256, dquant[1], IFIX(0.5), max[4]);
			idct(dcts + 320, out + 320, dquant[2], IFIX(0.5), max[5]);
			/* Up to here we are fine, performance wise. This 
			 * color space conversion is dead slow and is killing
			 * performance for large resolutions. It should not be
			 * necessary, as v4l provides support for planar yuv.
			 * */
			col221111(out, pic + (my * 16 * mcusx + mx) * 16 * bpp, mcusx * 16 * bpp);
		}
	}
	
	return;
}

/****************************************************************/
/**************       huffman decoder             ***************/
/****************************************************************/

static int fillbits (struct in *, int, unsigned int);
static int dec_rec2 (struct in *, struct dec_hufftbl *, int *, int, int);

static int fillbits(struct in *in, int le, unsigned int bi)
{
	int b;

	if (in->marker) {
		if (le <= 16)
			in->bits = bi << 16, le += 16;
		return le;
	}
	while (le <= 24) {
		b = *in->p++;
		bi = bi << 8 | b;
		le += 8;
	}
	in->bits = bi;		/* tmp... 2 return values needed */
	return le;
}

#define LEBI_GET(in)	(le = in->left, bi = in->bits)
#define LEBI_PUT(in)	(in->left = le, in->bits = bi)

#define GETBITS(in, n) (					\
  (le < (n) ? le = fillbits(in, le, bi), bi = in->bits : 0),	\
  (le -= (n)),							\
  bi >> le & ((1 << (n)) - 1)					\
)

#define UNGETBITS(in, n) (	\
  le += (n)			\
)


static int dec_rec2(struct in *in, struct dec_hufftbl *hu, int *runp, int c, int i)
{
	int le, bi;

	le = in->left;
	bi = in->bits;
	if (i) {
		UNGETBITS(in, i & 127);
		*runp = i >> 8 & 15;
		i >>= 16;
	} else {
		for (i = DECBITS; (c = ((c << 1) | GETBITS(in, 1))) >= (hu->maxcode[i]); i++);
		if (i >= 16) {
			in->marker = M_BADHUFF;
			return 0;
		}
		i = hu->vals[hu->valptr[i] + c - hu->maxcode[i - 1] * 2];
		*runp = i >> 4;
		i &= 15;
	}
	if (i == 0) {		/* sigh, 0xf0 is 11 bit */
		LEBI_PUT(in);
		return 0;
	}
	/* receive part */
	c = GETBITS(in, i);
	if (c < (1 << (i - 1)))
		c += (-1 << i) + 1;
	LEBI_PUT(in);
	return c;
}

#define DEC_REC(in, hu, r, i)	 (	\
  r = GETBITS(in, DECBITS),		\
  i = hu->llvals[r],			\
  i & 128 ?				\
    (					\
      UNGETBITS(in, i & 127),		\
      r = i >> 8 & 15,			\
      i >> 16				\
    )					\
  :					\
    (					\
      LEBI_PUT(in),			\
      i = dec_rec2(in, hu, &r, r, i),	\
      LEBI_GET(in),			\
      i					\
    )					\
)

inline static void decode_mcus(struct in *in, int *dct, int n, struct scan *sc, int *maxp)
{
	struct dec_hufftbl *hu;
	int i, r, t;
	int le, bi;

	memset(dct, 0, n * 64 * sizeof(*dct));
	le = in->left;
	bi = in->bits;

	while (n-- > 0) {
		hu = sc->hudc.dhuff;
		*dct++ = (sc->dc += DEC_REC(in, hu, r, t));

		hu = sc->huac.dhuff;
		i = 63;
		while (i > 0) {
			t = DEC_REC(in, hu, r, t);
			if (t == 0 && r == 0) {
				dct += i;
				break;
			}
			dct += r;
			*dct++ = t;
			i -= r + 1;
		}
		*maxp++ = 64 - i;
		if (n == sc->next)
			sc++;
	}
	LEBI_PUT(in);
}

static void dec_makehuff(struct dec_hufftbl *hu, int *hufflen, unsigned char *huffvals)
{
	int code, k, i, j, d, x, c, v;
	for (i = 0; i < (1 << DECBITS); i++)
		hu->llvals[i] = 0;

/*
 * llvals layout:
 *
 * value v already known, run r, backup u bits:
 *  vvvvvvvvvvvvvvvv 0000 rrrr 1 uuuuuuu
 * value unknown, size b bits, run r, backup u bits:
 *  000000000000bbbb 0000 rrrr 0 uuuuuuu
 * value and size unknown:
 *  0000000000000000 0000 0000 0 0000000
 */
	code = 0;
	k = 0;
	for (i = 0; i < 16; i++, code <<= 1) {	/* sizes */
		hu->valptr[i] = k;
		for (j = 0; j < hufflen[i]; j++) {
			hu->vals[k] = *huffvals++;
			if (i < DECBITS) {
				c = code << (DECBITS - 1 - i);
				v = hu->vals[k] & 0x0f;	/* size */
				for (d = 1 << (DECBITS - 1 - i); --d >= 0;) {
					if (v + i < DECBITS) {	/* both fit in table */
						x = d >> (DECBITS - 1 - v -
							  i);
						if (v && x < (1 << (v - 1)))
							x += (-1 << v) + 1;
						x = x << 16 | (hu-> vals[k] & 0xf0) << 4 |
							(DECBITS - (i + 1 + v)) | 128;
					} else
						x = v << 16 | (hu-> vals[k] & 0xf0) << 4 |
						        (DECBITS - (i + 1));
					hu->llvals[c | d] = x;
				}
			}
			code++;
			k++;
		}
		hu->maxcode[i] = code;
	}
	hu->maxcode[16] = 0x20000;	/* always terminate decode */
}

/****************************************************************/
/**************             idct                  ***************/
/****************************************************************/

#define ONE ((long)IFIX(1.))
#define S2  ((long)IFIX(0.382683432))
#define C2  ((long)IFIX(0.923879532))
#define C4  ((long)IFIX(0.707106781))

#define S22 ((long)IFIX(2 * 0.382683432))
#define C22 ((long)IFIX(2 * 0.923879532))
#define IC4 ((long)IFIX(1 / 0.707106781))

#define C3IC1 ((long)IFIX(0.847759065))	/* c3/c1 */
#define C5IC1 ((long)IFIX(0.566454497))	/* c5/c1 */
#define C7IC1 ((long)IFIX(0.198912367))	/* c7/c1 */

#define XPP(a,b) (t = a + b, b = a - b, a = t)
#define XMP(a,b) (t = a - b, b = a + b, a = t)
#define XPM(a,b) (t = a + b, b = b - a, a = t)

#define ROT(a,b,s,c) (	t = IMULT(a + b, s),	\
			a = IMULT(a, c - s) + t,	\
			b = IMULT(b, c + s) - t)

#define IDCT		\
(			\
  XPP(t0, t1),		\
  XMP(t2, t3),		\
  t2 = IMULT(t2, IC4) - t3,	\
  XPP(t0, t3),		\
  XPP(t1, t2),		\
  XMP(t4, t7),		\
  XPP(t5, t6),		\
  XMP(t5, t7),		\
  t5 = IMULT(t5, IC4),	\
  ROT(t4, t6, S22, C22),\
  t6 -= t7,		\
  t5 -= t6,		\
  t4 -= t5,		\
  XPP(t0, t7),		\
  XPP(t1, t6),		\
  XPP(t2, t5),		\
  XPP(t3, t4)		\
)

static unsigned char zig2[64] = {
	0, 2, 3, 9, 10, 20, 21, 35,
	14, 16, 25, 31, 39, 46, 50, 57,
	5, 7, 12, 18, 23, 33, 37, 48,
	27, 29, 41, 44, 52, 55, 59, 62,
	15, 26, 30, 40, 45, 51, 56, 58,
	1, 4, 8, 11, 19, 22, 34, 36,
	28, 42, 43, 53, 54, 60, 61, 63,
	6, 13, 17, 24, 32, 38, 47, 49
};

void idct( int *in, int *out, int *quant, long off, int max)
{
	long t0, t1, t2, t3, t4, t5, t6, t7, t ;
	long tmp[64], *tmpp;
	int i, j ,te;
	unsigned char *zig2p;
	
	t0 = off;
	if (max == 1) {
		t0 += in[0] * quant[0];
		for (i = 0; i < 64; i++)
			out[i] = ITOINT(t0);
		return;
	}
	zig2p = zig2;
	tmpp = tmp;
	for (i = 0; i < 8; i++) {
		 j = *zig2p++;
     	 	t0 += in[j] *(long) quant[j];
     		 j = *zig2p++;
      		t5 = in[j] * (long)quant[j];
      		j = *zig2p++;
      		t2 = in[j] * (long)quant[j];
      		j = *zig2p++;
      		t7 = in[j] * (long)quant[j];
      		j = *zig2p++;
      		t1 = in[j] * (long)quant[j];
      		j = *zig2p++;
      		t4 = in[j] * (long)quant[j];
      		j = *zig2p++;
      		t3 = in[j] * (long)quant[j];
      		j = *zig2p++;
      		t6 = in[j] * (long)quant[j];
		
		
		if ( (t1 | t2 | t3 | t4 | t5 | t6 | t7) == 0 ){
		
		tmpp[0 * 8] = t0;
		tmpp[1 * 8] = t0;
		tmpp[2 * 8] = t0;
		tmpp[3 * 8] = t0;
		tmpp[4 * 8] = t0;
		tmpp[5 * 8] = t0;
		tmpp[6 * 8] = t0;
		tmpp[7 * 8] = t0;
		
		tmpp++;
		t0 = 0;
		continue;
		}
		IDCT;
		tmpp[0 * 8] = t0;
		tmpp[1 * 8] = t1;
		tmpp[2 * 8] = t2;
		tmpp[3 * 8] = t3;
		tmpp[4 * 8] = t4;
		tmpp[5 * 8] = t5;
		tmpp[6 * 8] = t6;
		tmpp[7 * 8] = t7;
		tmpp++;
		t0 = 0;
	}
	for (i = 0,j = 0; i < 8; i++) {
		t0 = tmp[j + 0];
		t1 = tmp[j + 1];
		t2 = tmp[j + 2];
		t3 = tmp[j + 3];
		t4 = tmp[j + 4];
		t5 = tmp[j + 5];
		t6 = tmp[j + 6];
		t7 = tmp[j + 7];
		if ( (t1 | t2 | t3 | t4 | t5 | t6 | t7) == 0 ){
		te = ITOINT(t0);
		out[j + 0] = te;
		out[j + 1] = te;
		out[j + 2] = te;
		out[j + 3] = te;
		out[j + 4] = te; 
		out[j + 5] = te;
		out[j + 6] = te;
		out[j + 7] = te;
		j+=8;
		continue;
		}
		IDCT;
		out[j + 0] = ITOINT(t0);
		out[j + 1] = ITOINT(t1);
		out[j + 2] = ITOINT(t2);
		out[j + 3] = ITOINT(t3);
		out[j + 4] = ITOINT(t4);
		out[j + 5] = ITOINT(t5);
		out[j + 6] = ITOINT(t6);
		out[j + 7] = ITOINT(t7);
		j +=8;
	}
	
}

static unsigned char zig[64] = {
	0, 1, 5, 6, 14, 15, 27, 28,
	2, 4, 7, 13, 16, 26, 29, 42,
	3, 8, 12, 17, 25, 30, 41, 43,
	9, 11, 18, 24, 31, 40, 44, 53,
	10, 19, 23, 32, 39, 45, 52, 54,
	20, 22, 33, 38, 46, 51, 55, 60,
	21, 34, 37, 47, 50, 56, 59, 61,
	35, 36, 48, 49, 57, 58, 62, 63
};

static int aaidct[8] = {
	IFIX(0.3535533906), IFIX(0.4903926402),
	IFIX(0.4619397663), IFIX(0.4157348062),
	IFIX(0.3535533906), IFIX(0.2777851165),
	IFIX(0.1913417162), IFIX(0.0975451610)
};


inline static void idctqtab(unsigned char *qin, int *qout)
{
	int i, j;

	for (i = 0; i < 8; i++)
		for (j = 0; j < 8; j++)
			qout[zig[i * 8 + j]] = qin[zig[i * 8 + j]] * 
			  			IMULT(aaidct[i], aaidct[j]);
}

inline static void scaleidctqtab(int *q, int sc)
{
	int i;

	for (i = 0; i < 64; i++)
		q[i] = IMULT(q[i], sc);
}

/****************************************************************/
/**************          color decoder            ***************/
/****************************************************************/

#define ROUND

/*
 * YCbCr Color transformation:
 *
 * y:0..255   Cb:-128..127   Cr:-128..127
 *
 *      R = Y                + 1.40200 * Cr
 *      G = Y - 0.34414 * Cb - 0.71414 * Cr
 *      B = Y + 1.77200 * Cb
 *
 * =>
 *      Cr *= 1.40200;
 *      Cb *= 1.77200;
 *      Cg = 0.19421 * Cb + .50937 * Cr;
 *      R = Y + Cr;
 *      G = Y - Cg;
 *      B = Y + Cb;
 *
 * =>
 *      Cg = (50 * Cb + 130 * Cr + 128) >> 8;
 */

inline static void initcol(int q[][64])
{
	scaleidctqtab(q[1], IFIX(1.77200));
	scaleidctqtab(q[2], IFIX(1.40200));
}

/* This is optimized for the stupid sun SUNWspro compiler. */
#define STORECLAMP(a,x)				\
(						\
  (a) = (x),					\
  (unsigned int)(x) >= 256 ? 			\
    ((a) = (x) < 0 ? 0 : 255)			\
  :						\
    0						\
)

#ifdef ROUND

#define CBCRCG(xin)			\
(						\
  cb = outc[0 +xin],			\
  cr = outc[64+xin],			\
  cg = (50 * cb + 130 * cr + 128) >> 8		\
)

#else

#define CBCRCG(xin)			\
(						\
  cb = outc[0 +xin],			\
  cr = outc[64+xin],			\
  cg = (3 * cb + 8 * cr) >> 4			\
)

#endif

#if (SWAP_RGB == 1)
#define PIC(yin, xin, p, xout)			\
(						\
  y = outy[(yin) * 8 + xin],			\
  STORECLAMP(p[(xout) * bpp + 2], y + cr),	\
  STORECLAMP(p[(xout) * bpp + 1], y - cg),	\
  STORECLAMP(p[(xout) * bpp + 0], y + cb)		\
)

#else /* (SWAP_RGB == 1) */

#define PIC(yin, xin, p, xout)			\
(						\
  y = outy[(yin) * 8 + xin],			\
  STORECLAMP(p[(xout) * bpp + 0], y + cr),	\
  STORECLAMP(p[(xout) * bpp + 1], y - cg),	\
  STORECLAMP(p[(xout) * bpp + 2], y + cb)		\
)

#endif /* (SWAP_RGB == 1) */

static void col221111(int *out, unsigned char *pic, int width)
{
	int i, j, k;
	unsigned char *pic0, *pic1;
	int *outy, *outc;
	int cr, cg, cb, y;

	pic0 = pic;
	pic1 = pic + width;
	outy = out;
	outc = out + 64 * 4;
	for (i = 2; i > 0; i--) {
		for (j = 4; j > 0; j--) {
			for (k = 0; k < 8; k++) {
				CBCRCG(k);
				PIC(k/4*8+0, (k & 3) * 2 + 0, pic0, k * 2 + 0);
				PIC(k/4*8+0, (k & 3) * 2 + 1, pic0, k * 2 + 1);
				PIC(k/4*8+1, (k & 3) * 2 + 0, pic1, k * 2 + 0);
				PIC(k/4*8+1, (k & 3) * 2 + 1, pic1, k * 2 + 1);
			}
			outc += 8;
			outy += 16;
			pic0 += 2 * width;
			pic1 += 2 * width;
		}
		outy += 64 * 2 - 16 * 4;
	}
}

void init_nolibjpeg_decoder() 
{
	unsigned int   i, j, k, l;
	unsigned int   qIndex = 2;
	int            tc, th, tt, tac, tdc;
	const unsigned char *ptr;

	/* set up a quantization table */
	for (i = 0; i < 2; i++)
	{
		for (j = 0; j < 64; j++)
		{
			quant[i][j] = GsmartQTable[qIndex * 2 + i][j];
		}
	}

	/* set up the huffman table */
	ptr = (const unsigned char *) GsmartJPEGHuffmanTable;
	l   = GSMART_JPG_HUFFMAN_TABLE_LENGTH;
	while (l > 0)
	{
		unsigned int  hufflen[16];
		unsigned char huffvals[256];

		tc = *ptr++;
		th = tc & 15;
		tc >>= 4;
		tt = tc * 2 + th;
		if (tc > 1 || th > 1)
		{
			printf("died whilst setting up huffman table.\n");
			abort();
		}
		for (i = 0; i < 16; i++)
			hufflen[i] = *ptr++;
		l -= 1 + 16;
		k = 0;
		for (i = 0; i < 16; i++)
		{
			for (j = 0; j < hufflen[i]; j++)
				huffvals[k++] = *ptr++;
			l -= hufflen[i];
		}
		dec_makehuff(dhuff + tt, hufflen, huffvals);
	}

	/* set up the scan table */
	ptr = (const unsigned char *) GsmartJPEGScanTable;
	for (i = 0; i < 3; i++) {
		dscans[i].cid = *ptr++;
		tdc = *ptr++;
		tac = tdc & 15;
		tdc >>= 4;
		if (tdc > 1 || tac > 1)
		{
			printf("died whilst setting up scan table.\n");
			abort();
		}
		/* for each component */
		for (j = 0; j < 3; j++)
			if (comps[j].cid == dscans[i].cid)
				break;
		
		dscans[i].hv = comps[j].hv;
		dscans[i].tq = comps[j].tq;
		dscans[i].hudc.dhuff = dec_huffdc + tdc;
		dscans[i].huac.dhuff = dec_huffac + tac;
	}

	if (dscans[0].cid != 1 ||
	    dscans[1].cid != 2 ||
	    dscans[2].cid != 3)
	{
		printf("invalid cid found.\n");
		abort();
	}

	if (dscans[0].hv != 0x22 ||
	    dscans[1].hv != 0x11 ||
	    dscans[2].hv != 0x11)
	{
		printf("invalid hv found.\n");
		abort();
	}

	idctqtab(quant[dscans[0].tq], dquant[0]);
	idctqtab(quant[dscans[1].tq], dquant[1]);
	idctqtab(quant[dscans[2].tq], dquant[2]);
	initcol(dquant);
	
	dscans[0].next = 6 - 4;
	dscans[1].next = 6 - 4 - 1;
	dscans[2].next = 6 - 4 - 1 - 1;	/* 411 encoding */

}
